# ******************************************************************************
#  Copyright (c) 2021. Kneron Inc. All rights reserved.                        *
# ******************************************************************************

import os
import sys
import argparse
import time
import librosa
PWD = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(1, os.path.join(PWD, '..'))
sys.path.insert(1, os.path.join(PWD, '../example/'))

from utils.ExampleHelper import get_device_usb_speed_by_port_id
import kp
import cv2
import numpy as np
from transform import SignalTransform

SCPU_FW_PATH = os.path.join(PWD, '../../res/firmware/KL520/kdp2_fw_scpu.bin')
NCPU_FW_PATH = os.path.join(PWD, '../../res/firmware/KL520/kdp2_fw_ncpu.bin')

MODEL_FILE_PATH = os.path.join(PWD, '/home/pi/Desktop/nef/models_520.nef')
# MODEL_FILE_PATH = os.path.join(PWD, '../../res/models/KL520/resnet18_w224h224_kn-model-zoo/kl520_20001_resnet18_w224h224.nef')
IMAGE_FILE_PATH = os.path.join(PWD, '../../res/images/s.wav')
LOOP_TIME = 100


"""
post-process utils

modified from (path in toolchain docker): /workspace/ai_training/classification/inference.py
"""


def softmax(A):
    e = np.exp(A)
    return e / np.sum(e, keepdims=True)


def postprocess(pre_output):
    score = softmax(pre_output)
    return score


if __name__ == '__main__':
    
    parser = argparse.ArgumentParser(description='KL520 Kneron Model Zoo Generic Inference Example - Classification.')
    parser.add_argument('-p',
                        '--port_id',
                        help='Using specified port ID for connecting device (Default: port ID of first scanned Kneron '
                             'device)',
                        default=0,
                        type=int)
    parser.add_argument('-m',
                        '--model',
                        help='Model file path (.nef) (Default: {})'.format(MODEL_FILE_PATH),
                        default=MODEL_FILE_PATH,
                        type=str)
    parser.add_argument('-i',
                        '--img',
                        help='Image file path (Default: {})'.format(IMAGE_FILE_PATH),
                        default=IMAGE_FILE_PATH,
                        type=str)
    args = parser.parse_args()

    usb_port_id = args.port_id
    MODEL_FILE_PATH = args.model
    IMAGE_FILE_PATH = args.img

    """
    connect the device
    """
    try:
        print('[Connect Device]')
        device_group = kp.core.connect_devices(usb_port_ids=[usb_port_id])
        print(' - Success')
    except kp.ApiKPException as exception:
        print('Error: connect device fail, port ID = \'{}\', error msg: [{}]'.format(usb_port_id,
                                                                                     str(exception)))
        exit(0)

    """
    setting timeout of the usb communication with the device
    """
    print('[Set Device Timeout]')
    kp.core.set_timeout(device_group=device_group, milliseconds=5000)
    print(' - Success')

    """
    upload firmware to device
    """
    try:
        print('[Upload Firmware]')
        kp.core.load_firmware_from_file(device_group=device_group,
                                        scpu_fw_path=SCPU_FW_PATH,
                                        ncpu_fw_path=NCPU_FW_PATH)
        print(' - Success')
    except kp.ApiKPException as exception:
        print('Error: upload firmware failed, error = \'{}\''.format(str(exception)))
        exit(0)

    """
    upload model to device
    """
    try:
        print('[Upload Model]')
        model_nef_descriptor = kp.core.load_model_from_file(device_group=device_group,
                                                            file_path=MODEL_FILE_PATH)
        print(' - Success')
    except kp.ApiKPException as exception:
        print('Error: upload model failed, error = \'{}\''.format(str(exception)))
        exit(0)
    time_start = time.time()
    
    """
    prepare the image
    """
    print('[Read Image]')
    
    # img = cv2.imread(filename=IMAGE_FILE_PATH)

    #img_bgr565 = cv2.cvtColor(src=img, code=cv2.COLOR_BGR2BGR565)
    print(' - Success')

    """
    prepare app generic inference config
    """
    generic_raw_image_header = kp.GenericRawImageHeader(
        model_id=model_nef_descriptor.models[0].id,
        resize_mode=kp.ResizeMode.KP_RESIZE_ENABLE,
        padding_mode=kp.PaddingMode.KP_PADDING_CORNER,
        normalize_mode=kp.NormalizeMode.KP_NORMALIZE_KNERON,
        inference_number=0
    )

    """
    starting inference work
    """
    print('[Starting Inference Work]')
    print(' - Starting inference loop {} times'.format(LOOP_TIME))
    print(' - ', end='')
    #y, fs = librosa.core.load(IMAGE_FILE_PATH, sr=None, mono=True,res_type='kaiser_fast')
    #transforms = SignalTransform(sr=fs, n_mels=64, n_mfcc=40)
    for i in range(LOOP_TIME):
        y, fs = librosa.core.load(IMAGE_FILE_PATH, sr=None, mono=True,res_type='kaiser_fast')
        transforms = SignalTransform(sr=fs, n_mels=64, n_mfcc=40)
        len_y = len(y)
        # effective_length = self.sr * self.period
        effective_length = fs * 4
        if len_y < effective_length:
            new_y = np.zeros(effective_length, dtype=y.dtype)
            start = np.random.randint(effective_length - len_y)
            new_y[start:start + len_y] = y
            y = new_y.astype(np.float32)
        elif len_y > effective_length:
            start = np.random.randint(len_y - effective_length)
            y = y[start:start + effective_length].astype(np.float32)
        else:
            y = y.astype(np.float32)
        image = transforms.melspectransform(y)
        # image = cv2.resize(np.expand_dims(transforms.melspectransform(y).astype(np.uint8),axis=0),(224,224))
        image = np.expand_dims(image, axis=0)
        image = np.transpose(image, (1,2,0))
        image = np.expand_dims(image, axis=0)
        
        # img_bgr565 = cv2.cvtColor(src=image, code=cv2.COLOR_GRAY2BGR565)
        try:
            kp.inference.generic_raw_inference_send(device_group=device_group,
                                                    generic_raw_image_header=generic_raw_image_header,
                                                    image=image,
                                                    image_format=kp.ImageFormat.KP_IMAGE_FORMAT_RAW8)

            generic_raw_result = kp.inference.generic_raw_inference_receive(device_group=device_group,
                                                                            generic_raw_image_header=generic_raw_image_header,
                                                                            model_nef_descriptor=model_nef_descriptor)
            
        except kp.ApiKPException as exception:
            print(' - Error: inference failed, error = {}'.format(exception))
            exit(0)

        print('.', end='', flush=True)
    print()

    """
    retrieve inference node output 
    """
    print('[Retrieve Inference Node Output ]')
    inf_node_output_list = []
    for node_idx in range(generic_raw_result.header.num_output_node):
        inference_float_node_output = kp.inference.generic_inference_retrieve_float_node(node_idx=node_idx,
                                                                                         generic_raw_result=generic_raw_result,
                                                                                         channels_ordering=kp.ChannelOrdering.KP_CHANNEL_ORDERING_CHW)
        inf_node_output_list.append(inference_float_node_output)

    print(' - Success')

    print('[Result]')

    tmp = []
    for o_n in inf_node_output_list:
        o_array = o_n.ndarray.copy()
        tmp.append(o_array)

    res = postprocess(tmp)
    time_end = time.time()
    time_c = time_end - time_start
    print('time cost:',time_c)
    print("Top1 class: " + str(np.argmax(res)) + ", score: " + str(np.max(res)))
